# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from datetime import datetime

author = "tadams42"
project = "seveno-pyutil"
copyright = (
    ", ".join(str(y) for y in range(2017, datetime.now().year + 1)) + ", " + author
)
release = "0.8.4"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "myst_parser",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]

source_suffix = {".rst": "restructuredtext", ".md": "markdown"}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.9", None),
    "marshmallow": ("https://marshmallow.readthedocs.io/en/3.0/", None),
    "json": ("https://simplejson.readthedocs.io/en/latest/", None),
    "simplejson": ("https://simplejson.readthedocs.io/en/latest/", None),
}

add_module_names = False


# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False



import logging
import sys
from logging import Logger, NullHandler
from pathlib import Path
from typing import Union


def silence_logger(logger: Logger):
    """
    For given logger, replaces all its handlers with :class:`logging.NullHandler`.
    """
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])

    logger.addHandler(NullHandler())


def log_to_console_for(logger_name: str):
    """
    Sometimes, usually during development, we want to quickly see output of some
    particular package logger. This method will configure such logger to spit stuff out
    to console.
    """
    logger = logging.getLogger(logger_name)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)


def log_to_tmp_file_for(
    logger_name: str, file_path: Union[str, Path] = "/tmp/seveno_pyutil.log"
):
    """
    Quick setup for given logger directing it to ``/tmp/seveno_pyutil.log`` This is of
    course mainly used during development, especially when playing with things in Python
    console.
    """
    logger = logging.getLogger(logger_name)
    handler = logging.FileHandler(filename=file_path)
    handler.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)



from datetime import date, datetime, timedelta
from zoneinfo import ZoneInfo

from seveno_pyutil import ensure_tzinfo
from seveno_pyutil.datetime_utilities import iter_year_month


class describe_ensure_tzinfo(object):
    def it_converts_naive_datetime_to_timezone_aware_datetime(self):
        dt = datetime.now()
        assert ensure_tzinfo(dt, tz_or_offset=None).tzinfo is ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="Z").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="UTC").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset=ZoneInfo("UTC")).tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="-02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02").tzinfo is not None
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=2, minutes=30)).tzinfo
            is not None
        )
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=-2, minutes=30)).tzinfo
            is not None
        )

    def it_converts_timezone_aware_datetime_to_different_timezone(self):
        dt = datetime.now().replace(tzinfo=ZoneInfo("Europe/Zagreb"))

        assert ensure_tzinfo(dt, tz_or_offset=None).tzinfo is ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="Z").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="UTC").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset=ZoneInfo("UTC")).tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="-02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02").tzinfo is not None
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=2, minutes=30)).tzinfo
            is not None
        )
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=-2, minutes=30)).tzinfo
            is not None
        )


class describe_iter_year_month:
    def it_generates_empty_range_when_start_is_after_end(self):
        start = date(2022, 12, 1)
        end = date(2022, 9, 1)

        assert not list(iter_year_month(start, end))
        assert not list(iter_year_month(start, end, include_start=True))
        assert not list(iter_year_month(start, end, include_end=True))
        assert not list(
            iter_year_month(start, end, include_start=True, include_end=True)
        )

    def it_generates_empty_range_when_start_is_equal_to_end_and_borders_are_excluded(
        self,
    ):
        start = date(2022, 12, 1)
        end = date(2022, 12, 1)
        assert not list(iter_year_month(start, end))

    def it_generated_single_entry_when_start_is_equal_to_end_and_any_of_borders_is_included(
        self,
    ):
        start = date(2022, 12, 1)
        end = date(2022, 12, 1)
        assert len(list(iter_year_month(start, end, include_start=True))) == 1
        assert len(list(iter_year_month(start, end, include_end=True))) == 1
        assert (
            len(list(iter_year_month(start, end, include_start=True, include_end=True)))
            == 1
        )

    def it_generates_date_range(self):
        start = date(2022, 9, 15)
        end = date(2023, 3, 31)

        assert list(iter_year_month(start, end)) == [
            date(2022, 10, 1),
            date(2022, 11, 1),
            date(2022, 12, 1),
            date(2023, 1, 1),
            date(2023, 2, 1),
        ]

    def it_may_include_borders(self):
        start = date(2022, 9, 15)
        end = date(2023, 3, 31)

        assert list(iter_year_month(start, end, include_start=True)) == [
            date(2022, 9, 1),
            date(2022, 10, 1),
            date(2022, 11, 1),
            date(2022, 12, 1),
            date(2023, 1, 1),
            date(2023, 2, 1),
        ]

        assert list(iter_year_month(start, end, include_end=True)) == [
            date(2022, 10, 1),
            date(2022, 11, 1),
            date(2022, 12, 1),
            date(2023, 1, 1),
            date(2023, 2, 1),
            date(2023, 3, 1),
        ]

        assert list(
            iter_year_month(start, end, include_start=True, include_end=True)
        ) == [
            date(2022, 9, 1),
            date(2022, 10, 1),
            date(2022, 11, 1),
            date(2022, 12, 1),
            date(2023, 1, 1),
            date(2023, 2, 1),
            date(2023, 3, 1),
        ]



from seveno_pyutil import is_blank


class DescribeIsBlank(object):
    def it_works_for_strings_numbers_and_iterables(self):
        assert is_blank("") == True
        assert is_blank("     ") == True
        assert is_blank("  \t \n \r   ") == True
        assert is_blank(None) == True
        assert is_blank(0) == True
        assert is_blank([]) == True
        assert is_blank([None, None]) == True

        assert is_blank(42) == False
        assert is_blank("42") == False
        assert is_blank([None, 42]) == False



import re
from datetime import date, datetime, timedelta, timezone, tzinfo
from typing import Optional, Union
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import holidays

TimeZoneLike = Union[str, int, timedelta, tzinfo, ZoneInfo, timezone]

CROATIAN_HOLIDAYS = holidays.HR()
_ONE_DAY = timedelta(days=1)
_ISO_8601_OFFSET = re.compile(r"([+-]?)([0-9]{2})[:]?([0-9]{0,2})")


def next_working_day(from_: Optional[date] = None, holidays_calendar=CROATIAN_HOLIDAYS):
    """Finds next work day from `from_` or today."""
    next_day = (from_ or date.today()) + _ONE_DAY
    while next_day.weekday() in holidays.WEEKEND or next_day in holidays_calendar:
        next_day += _ONE_DAY
    return next_day


def timezone_or_offset(from_: Optional[TimeZoneLike]) -> Union[ZoneInfo, timezone]:
    """
    Given ``from_`` creates `datetime.timezone` or `zoneinfo.ZoneInfo` as
    result.

    ``from_`` can be any of following:

    - `str` (ie. "-02:42", None, "", "Z", ...) which is ISO8601 offset
    - `str` (ie. "Europe/Zagreb") which is timezone name
    - `int` (ie. -9000) which is total number of seconds in time offset
    - `datetime.timedelta`
    - `datetime.tzinfo` or something that behaves like it
    """
    offset_obj = None

    if from_ is None:
        offset_obj = ZoneInfo("UTC")

    elif isinstance(from_, (ZoneInfo, timezone)):
        return from_

    elif isinstance(from_, str):
        try:
            offset_obj = ZoneInfo(from_)
        except (ZoneInfoNotFoundError, ValueError):
            pass

        if not offset_obj:
            if from_.strip() in ["Z", ""]:
                offset_obj = ZoneInfo("UTC")
            else:
                match_object = _ISO_8601_OFFSET.match(from_)
                if match_object:
                    sign, hours, minutes = match_object.groups()
                    offset_obj = timezone(
                        name=from_,
                        offset=timedelta(
                            hours=(-1 if sign == "-" else 1) * int(hours),
                            minutes=int(minutes or 0),
                        ),
                    )

    elif isinstance(from_, timedelta):
        offset_obj = timezone(name="{} s".format(from_.total_seconds()), offset=from_)

    elif isinstance(from_, int):
        offset_obj = timezone(
            name="{} s".format(from_), offset=timedelta(seconds=from_)
        )

    elif issubclass(type(from_), tzinfo) or all(
        hasattr(from_, attr_name)
        for attr_name in ["dst", "fromutc", "tzname", "utcoffset"]
    ):
        offset_obj = timezone(name=from_.tzname(None), offset=from_.utcoffset(None))

    if offset_obj is None:
        raise ValueError("Unable to parse time offset: {}".format(from_))

    return offset_obj


def ensure_tzinfo(
    val: datetime, tz_or_offset: TimeZoneLike = "UTC", is_dst: bool = False
) -> datetime:
    """
    Creates timezone aware datetime object for ``val``.

    - if ``val`` is naive datetime, new value will be created as datetime
      localized in ``tz_or_offset`` timezone
    - if ``val`` is already timezone aware, it will be converted to
      ``tz_or_offset`` timezone using `datetime.datetime.astimezone`

    Arguments:
        val: Input value for conversion
        tz_or_offset: Anything that `timezone_or_offset` accepts
        is_dst: used to determine the correct timezone in the ambiguous
            period at the end of daylight saving time. Use ``is_dst=None`` to
            raise an AmbiguousTimeError for ambiguous times at the end of
            daylight saving time.

    Return:
        Timezone aware datetime object

    Raises:
        ValueError: When timezone of offset can't be parsed / determined from
            ``tz_or_offset``

    Note:
        This tries to provide safe(ish) implementation for handling naive
        datetime objects, but ultimate solution is to not use naive datetime
        objects ever/anywhere. Recommendation is to go with ``pip install
        pendulum`` and leave this crap behind to history.
    """
    if not isinstance(val, datetime):
        raise ValueError(
            "Input is not datetime! ensure_tzinfo doesn't parse datetime, you "
            "need to do that beforehand."
        )

    tz_or_offset = timezone_or_offset(tz_or_offset)

    if not val.tzinfo:
        val = val.replace(tzinfo=tz_or_offset)
    else:
        val = val.astimezone(tz_or_offset)

    return val


def iter_year_month(
    start: Union[date, datetime],
    end: Optional[date] = None,
    include_start: bool = False,
    include_end: bool = False,
):
    """
    Generates range of date(year, month, 1) from ``start`` to ``end``.

    - when ``start > end`` generated range is empty
    - when ``start == end`` generated range may contain single ``date`` object
      depending on ``include_start`` or ``include_end``

    Arguments:
        start: begin of range
        end: end of range. If ``None``, assumes, ``start == end``
        include_start: include ``start`` in generated range
        include_end: include ``end`` in generated range
    """

    if not end:
        end = start

    start = date(year=start.year, month=start.month, day=1)
    end = date(year=end.year, month=end.month, day=1)

    if start > end:
        return

    if start == end:
        if include_start or include_end:
            yield start
        return

    val = date(year=start.year, month=start.month, day=1)

    if include_start:
        yield val

    end_reached = False
    while not end_reached:
        try:
            val = val.replace(month=val.month + 1)

        except ValueError:
            if val.month == 12:
                val = val.replace(year=val.year + 1, month=1)

        if val == end:
            if include_end:
                yield val
            end_reached = True

        else:
            yield val



import logging

import colorlog
from pretty_traceback.formatting import exc_to_traceback_str


class PrettyFormatter(colorlog.ColoredFormatter):
    """
    Logging formatter for pretty logs:

    - can optionally colorize output
    - can optionally force output to be single line
    - reformats logged tracebacks

    To colorize output:

    - instantiate formatter with `colorize = True`
    - use color placeholders in format string, ie `%(red)s text that will be red %(reset)s`

    Example YAML logging config to use it:

    .. code-block:: yaml

        ---
        version: 1

        loggers:
          my_app:
            handlers:
              - console
            level: INFO
            propagate: false

        handlers:
          console:
            class: logging.StreamHandler
            filters:
              - request_id
            stream: ext://sys.stdout
            formatter: colored_multiline

        formatters:
          colored_multiline:
            (): seveno_pyutil.PrettyFormatter
            force_single_line: false
            colorize: true
            format: >-
              lvl=%(log_color)s%(levelname)s%(reset)s
              ts=%(thin_white)s%(asctime)s%(reset)s
              msg=%(message)s

          colorless_single_line:
            (): seveno_pyutil.PrettyFormatter
            force_single_line: true
            colorize: false
            format: >-
              lvl=%(levelname)s
              ts=%(asctime)s
              msg=%(message)s

    Some of available colors are:

    .. code-block:: python

        [
            'black', 'bold_black', 'thin_black', 'red', 'bold_red', 'thin_red', 'green',
            'bold_green', 'thin_green', 'yellow', 'bold_yellow', 'thin_yellow', 'blue',
            'bold_blue', 'thin_blue', 'purple', 'bold_purple', 'thin_purple', 'cyan',
            'bold_cyan', 'thin_cyan', 'white', 'bold_white', 'thin_white', "..."
        ]

    Others can be retrieved via:

    .. code-block:: python

        colorlog.ColoredFormatter()._escape_code_map("DEBUG").keys()

    or check docs for `colorlog`.
    """

    def __init__(
        self, *args, force_single_line: bool = True, colorize: bool = False, **kwargs
    ):
        self.force_single_line = force_single_line
        self.colorize = colorize

        # Adjust any base class arguments that might had slipped in
        kwargs.pop("force_color", None)
        kwargs.pop("no_color", None)
        if not self.colorize:
            kwargs["no_color"] = True
        else:
            kwargs["force_color"] = True

        super().__init__(*args, **kwargs)

    def formatException(self, ei) -> str:
        _, exc_value, traceback = ei
        return exc_to_traceback_str(exc_value, traceback, color=self.colorize)

    def format(self, record: logging.LogRecord) -> str:
        retv = super().format(record)

        if self.force_single_line:
            return retv.replace("\n", "\\n")

        return retv



from __future__ import annotations

import enum
import json
import logging
import timeit
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING, Callable, ClassVar, Final
from uuid import UUID

import pygments
import sqlparse

# from pygments.formatters import TerminalTrueColorFormatter
from pygments.formatters import Terminal256Formatter
from pygments.lexers import SqlLexer

HAS_FLASK = False
try:
    import flask

    HAS_FLASK = True
except Exception:
    pass

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection, Engine


class SQLFilter(logging.Filter):
    """
    Filter for SQLAlchemy SQL loggers. Optionally reformats and colorizes queries.

    To use it with SQLAlchemy:

    - add it to ``sqlalchemy.engine`` logger and call
      `:meth:register_sqlalchemy_logging_events`

    Supports following logging placeholders:

    +-------------------+----------------------------------------------+
    | placeholder       | description                                  |
    +-------------------+----------------------------------------------+
    | %(sql)s           | Formatted SQL statement that was executed    |
    +-------------------+----------------------------------------------+
    | %(sql_duration)s  | Formatted duration of SQL execution          |
    +-------------------+----------------------------------------------+

    Arguments:
        colorize_queries(bool): Should apply shell coloring escape sequences to
            formatted SQL?
        multiline_queries(bool): Should emit SQL as indented, multiline of
            single line log statements? In development it is usually nice to
            have it be `True`. In production environments, multiline logs are
            pain and should be avoided.

    Example::

        import logging
        from logging.config import dictConfig

        from seveno_pyutil import SQLFilter

        try:
            import sqlalchemy
            SQLFilter.register_sqlalchemy_logging_events('myapp.db')
        except ImportError:
            pass

        dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'sqlalchemy_sql': {'format': '(%(sql_duration)s) %(sql)s %(message)s'}
            },
            'filters': {
                'colored_sql': {
                    '()': 'seveno_pyutil.SQLFilter'
                    'colorize_queries': True,
                    'multiline_queries': True,
                }
            },
            'handlers': {
                'console_sqlalchemy': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'sqlalchemy_sql',
                    'filters': ['colored_sql'],
                    'stream': 'ext://sys.stdout'
                }
            },
            'loggers': {
                'myapp.db': {
                    'level': 'DEBUG',
                    'propagate': False,
                    'handlers': ['console_sqlalchemy']
                }
            }
        })

    """

    @classmethod
    def register_sqlalchemy_logging_events(
        cls,
        logger: str | logging.Logger,
        duration_threshold_ms: int | float | timedelta | None = None,
    ):
        """
        Must be called when logging SQLAlchemy statements and durations.

        Arguments:
            logger: name of logger or logging.Logger instance that will be used to log
                SQL messages.
            duration_threshold_ms: If given, only queries lasting longer than this
                threshold will be logged.
            statistics_ctx_get: callable that returns mutable dict. If given, then each
                time SQL query is executes, this dict will be updated.

                It can be used for example in Flask to track SQL execution statistics
                during one HTTP request::

                    import flask
                    from seveno_pyutil import FlaskSQLStats, SQLFilter

                    SQLFilter.register_sqlalchemy_logging_events(
                        "may_app_sql_logger", 100
                    )

                    app = flask.Flask()

                    @app.before_request
                    def log_current_request():
                        FlaskSQLStats.open()

                        @flask.after_this_request
                        def log_current_response(response):
                            FlaskSQLStats.close()
                            return response
        """

        from sqlalchemy import event
        from sqlalchemy.engine import Engine

        connection_enricher = ConnectionEnricher(logger, duration_threshold_ms)

        @event.listens_for(Engine, "before_cursor_execute")
        def before_cursor_execute(
            conn: Connection,
            cursor: "DBAPICursor",
            statement: str,
            parameters: dict,
            context: "ExecutionContext",
            executemany: bool,
        ):
            connection_enricher.start(conn)

        @event.listens_for(Engine, "after_cursor_execute")
        def after_cursor_execute(
            conn: Connection,
            cursor: "DBAPICursor",
            statement: str,
            parameters: dict,
            context: "ExecutionContext",
            executemany: bool,
        ):
            connection_enricher.end(conn, cursor, statement, parameters)

        @event.listens_for(Engine, "handle_error")
        def receive_handle_error(exception_context):
            connection_enricher.error(exception_context)

    def __init__(
        self,
        colorize_queries=False,
        multiline_queries=False,
        shorten_logs=True,
        *args,
        **kwargs,
    ):
        self.enricher = RecordEnricher(
            colorize_queries, multiline_queries, shorten_logs
        )
        super().__init__(*args, **kwargs)

    def filter(self, record: logging.LogRecord):
        self.enricher.add_attributes(record)
        return super().filter(record)


class ConnectionEnricher:
    _ATTR_START_TIME = "query_start_time"

    def __init__(
        self,
        lgr: str | logging.Logger,
        duration_threshold: int | float | timedelta | None,
    ):
        self.lgr: logging.Logger
        if isinstance(lgr, str):
            self.lgr = logging.getLogger(lgr)
        else:
            self.lgr = lgr

        self.duration_threshold: timedelta | None = None
        if duration_threshold is not None:
            if isinstance(duration_threshold, timedelta):
                self.duration_threshold = duration_threshold
            else:
                self.duration_threshold = timedelta(milliseconds=duration_threshold)

    def start(self, conn: Connection | None):
        if conn:
            conn.info.setdefault(self._ATTR_START_TIME, []).append(
                timeit.default_timer()
            )

    def end(
        self, conn: Connection, cursor: "DBAPICursor", statement: str, parameters: dict
    ):
        statement_duration = self._execution_duration(conn)
        FlaskSQLStats.incr_stats(self.lgr, statement_duration)

        compiled = None
        try:
            compiled = cursor.mogrify(statement, parameters).decode()
        except Exception:
            compiled = None

        warned = False
        if (
            self.duration_threshold is not None
            and statement_duration >= self.duration_threshold
        ):
            self.lgr.warning(
                "Detected SQL query running longer than {:.2f} ms! ".format(
                    self.duration_threshold.total_seconds() * 1000
                ),
                extra={
                    RecordEnricher.ATTR_DATA: SQLRecordedQuery(
                        statement=str(statement),
                        parameters=parameters,
                        duration=statement_duration,
                        compiled=compiled,
                    )
                },
            )
            warned = True

        if not warned and statement_duration is not None:
            self.lgr.debug(
                "",
                extra={
                    RecordEnricher.ATTR_DATA: SQLRecordedQuery(
                        statement=str(statement),
                        parameters=parameters,
                        duration=statement_duration,
                        compiled=compiled,
                    )
                },
            )

    def error(self, exception_context):
        msg = "SQL engine exception detected."

        try:
            msg = msg + " " + str(exception_context.original_exception).strip()
        except Exception:
            msg = "SQL engine exception detected. "

        self.lgr.critical(
            msg,
            extra={
                RecordEnricher.ATTR_DATA: SQLRecordedQuery(
                    statement=str(exception_context.statement),
                    parameters=exception_context.parameters,
                    duration=self._execution_duration(exception_context.connection),
                )
            },
        )

    @classmethod
    def _execution_duration(cls, conn: Connection | None) -> timedelta:
        data: list[float] = getattr(conn, "info", dict()).get(cls._ATTR_START_TIME, [])
        started_at = data.pop(-1) if data else 0
        return timedelta(milliseconds=(timeit.default_timer() - started_at) * 1000.0)


@dataclass
class SQLRecordedQuery:
    statement: str
    parameters: dict
    duration: timedelta
    compiled: str | None = None
    duration_ms: float = field(init=False, default=0)

    def __post_init__(self):
        td = None
        if isinstance(self.duration, (int, float, str)):
            td = timedelta(microseconds=float(self.duration))
        elif isinstance(self.duration, timedelta):
            td = self.duration
        else:
            td = timedelta()
        self.duration_ms = td.total_seconds() * 1000


class RecordEnricher:
    ATTR_DATA = "_sql"
    ATTR_SQL = "sql"
    ATTR_DURATION = "sql_duration"

    def __init__(
        self, colorize_queries=False, multiline_queries=False, shorten_logs=True
    ):
        self.colorize_queries = colorize_queries
        self.multiline_queries = multiline_queries
        self.shorten_logs = shorten_logs

    def add_attributes(self, record: logging.LogRecord):
        setattr(record, self.ATTR_SQL, "")
        setattr(record, self.ATTR_DURATION, "")

        recorded_query: SQLRecordedQuery | None = getattr(record, self.ATTR_DATA, None)
        if not recorded_query:
            return

        query_str = self._format_query_string(recorded_query)
        setattr(record, self.ATTR_SQL, query_str)

        duration_str = self._format_duration_string(recorded_query)
        setattr(record, self.ATTR_DURATION, duration_str)

    def _format_query_string(self, recorded_query: SQLRecordedQuery) -> str:
        if recorded_query.compiled:
            sql = self._format_compiled(recorded_query.compiled)
            return sql

        else:
            sql = self._format_statement_and_params(
                recorded_query.statement, recorded_query.parameters
            )
            return sql

    _SQL_FORMAT_OPTS = {
        "reindent": True,
        "keyword_case": "upper",
        "truncate_strings": 25,
        "reindent_aligned": True,
        "wrap_after": 88,
    }

    def _maybe_multiline(self, sql: str) -> str:
        if sql:
            if self.multiline_queries:
                sql = "\n".join(
                    l.rstrip()
                    for l in sqlparse.format(sql, **self._SQL_FORMAT_OPTS).splitlines()
                    if l.strip()
                ).strip()
            else:
                sql = " ".join(
                    l.strip()
                    for l in sqlparse.format(sql, **self._SQL_FORMAT_OPTS).splitlines()
                    if l.strip()
                ).strip()

            if sql and not sql.endswith(";"):
                sql = sql + ";"

        return sql or ""

    def _maybe_colorized(self, sql: str) -> str:
        if self.colorize_queries and sql:
            sql = pygments.highlight(
                sql, SqlLexer(), Terminal256Formatter(style="monokai")
            ).strip()

        return sql or ""

    def _format_compiled(self, compiled: str | None) -> str:
        sql = compiled or ""

        sql = self._maybe_multiline(sql)
        sql = self._maybe_colorized(sql)

        if self.shorten_logs:
            (sql or " SQL")[:1300]

        return sql

    def _format_statement_and_params(
        self, statement: str | None, parameters: dict | None
    ) -> str:
        sql: str = statement or ""

        sql = self._maybe_multiline(sql)
        sql = self._maybe_colorized(sql)

        if parameters:
            params_dict = parameters
            params = json.dumps(params_dict, cls=JSONEncoder).strip()
        else:
            params_dict = {}
            params = ""

        if self.colorize_queries and params:
            params = pygments.highlight(
                params,
                pygments.lexers.get_lexer_for_mimetype("application/json"),
                Terminal256Formatter(style="monokai"),
            ).strip()

        # rsyslogd limits to 2048 bytes per message by default
        # We prefer to lose some params when shortening log line than to lose some SQL
        # statement content. Either way, it is not possible to guarantee both will
        # be fully logged in all contexts and log sinks
        if params and params_dict:
            if self.shorten_logs:
                sql = "{} with params {}".format(sql[:1500], params[:500])
            else:
                # params should always be shortened because they can be huge in when
                # for exmple we are inserting into PostgreSQL JSONB columns
                sql = "{} with params {}".format(sql, params[:500])
        else:
            sql = (sql or " SQL")[:1300]

        return sql

    def _format_duration_string(self, recorded_query: SQLRecordedQuery) -> str:
        dur = "_.___ ms"
        if recorded_query.duration_ms:
            dur = "{:.2f} ms".format(recorded_query.duration_ms)
        return dur


class FlaskSQLStats:
    _KEY_REQUEST_SQL_STATS = "sqlalchemy_statistics"
    _KEY_SQL_CUMULATIVE_DURATION = "cumulative_duration"
    _KEY_SQL_COUNT = "statements_count"

    @classmethod
    def get_statements_count(cls) -> int:
        stats = cls.get()
        if stats:
            return stats[cls._KEY_SQL_COUNT] or 0
        return 0

    @classmethod
    def get_cumulative_statements_duration(cls) -> timedelta:
        stats = FlaskSQLStats.get()
        if stats:
            return stats[cls._KEY_SQL_CUMULATIVE_DURATION]
        return timedelta()

    @classmethod
    def open(cls) -> dict | None:
        if HAS_FLASK:
            return flask.g.setdefault(cls._KEY_REQUEST_SQL_STATS, dict())

    @classmethod
    def close(cls):
        if HAS_FLASK:
            flask.g.pop(cls._KEY_REQUEST_SQL_STATS)

    @classmethod
    def get(cls) -> dict | None:
        if HAS_FLASK:
            return flask.g.get(cls._KEY_REQUEST_SQL_STATS)

    @classmethod
    def incr_stats(cls, lgr: logging.Logger, statement_duration: timedelta):
        if HAS_FLASK:
            try:
                data = cls.open()
                data.setdefault(cls._KEY_SQL_CUMULATIVE_DURATION, timedelta())
                data.setdefault(cls._KEY_SQL_COUNT, 0)
                data[cls._KEY_SQL_CUMULATIVE_DURATION] += statement_duration
                data[cls._KEY_SQL_COUNT] += 1

            except Exception:
                lgr.error("Failed to collect SQL statistics!", exc_info=True)


class JSONEncoder(json.JSONEncoder):
    _DATES_TIMES = (date, datetime)

    def default(self, obj):
        if isinstance(obj, self._DATES_TIMES):
            return obj.isoformat()
        elif isinstance(obj, enum.Enum):
            return obj.name
        elif isinstance(obj, (UUID, Decimal, Path)):
            return str(obj)
        # return json.JSONEncoder.default(self, obj)
        return str(obj)



from seveno_pyutil.file_utilities import switch_extension


class describe_switch_extension(object):
    def it_returns_file_path_with_new_file_extension(self):
        assert switch_extension("/foo/bar/baz.txt", "rtf") == "/foo/bar/baz.rtf"
        assert switch_extension("/foo/bar/baz.txt", ".rtf") == "/foo/bar/baz.rtf"

    def it_returns_same_file_path_if_given_extension_is_empty(self):
        assert switch_extension("/foo/bar/baz.txt", "") == "/foo/bar/baz.txt"



from collections import abc
from typing import Mapping, Sequence, Union

from .metaprogramming_helpers import getval
from .string_utilities import is_blank


class ExceptionsAsErrors:
    """
    Context manager that swallows exceptions and stores them as structured error dict
    that can later be added to `marshmallow.ValidationError`.

    It uses `add_error_to` to update provided ``errors_store`` with caught exceptions.

    Example::

        errors = {}
        with ExceptionsAsErrors(errors) as e:
            raise RuntimeError("ZOMG!")
        errors == {'_schema': ['ZOMG!']}

        errors = {}
        with ExceptionsAsErrors(errors, subkey="some_name") as e:
            raise RuntimeError("ZOMG!")
        errors == {'some_name': ['ZOMG!']}
    """

    def __init__(self, errors_store: Union[Mapping, object], subkey=None):
        self.subkey = subkey if not is_blank(subkey) else None
        self.errors_store = errors_store

    def __enter__(self):
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        if exc_val:
            if self.subkey:
                add_error_to(self.errors_store, {self.subkey: exc_val})
            else:
                add_error_to(self.errors_store, exc_val)
        return True


def _normalize_error_data(error):
    if isinstance(error, str):
        return error

    if isinstance(error, (abc.Sequence, abc.Set)):
        return [_normalize_error_data(obj) for obj in error]

    if isinstance(error, abc.Mapping):
        return {k: _normalize_error_data(v) for k, v in error.items()}

    if hasattr(error, "normalized_messages"):
        return _normalize_error_data(error.normalized_messages())

    return str(error)


def add_error_to(
    errors_store: Union[Mapping, object],
    error: Union[str, Sequence, object, Mapping, Exception],
):
    """
    Updates error store, merging messages from ``error``

    Example::

        errors_store = {
            "person": {
                "email": ["is not an email"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"]
            },
            "job": ["is not from allowed values list"]
        }

        add_error_to(errors_store, {"person": {"email": "is from illegal domain"}})

        errors_store == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"]
            },
            "job": ["is not from allowed values list"]
        }

    But, trying to do this::

        add_error_to(errors_store, {"person": "is illegally formed"})

    will add/update ``_schema`` key because ``person`` has child keys and replacing
    them with ``["is illegally formed"]`` would loose that data. Thus,
    ``errors_store`` will look like this now::

        errors_store == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"]
            },
            "job": ["is not from allowed values list"]
        }

    Another example of this behavior::

        add_error_to(errors_store, {"job": {"title": "can't be blank"}})

    will result in::

        errors_store == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"]
            },
            "job": {
                "_schema": ["is not from allowed values list"],
                "title": ["can't be blank"]
            }
        }

    Also, method is smart enough to correctly update errors store with either one or
    a sequence of messages. So both of these are valid::

        add_error_to(errors_store, {"job": {"title": "is overpaid"}})
        add_error_to(errors_store, {"job": {"title": ["is forbiden", "doesn't exist"]}})

        errors_store == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"]
            },
            "job": {
                "_schema": ["is not from allowed values list"],
                "title": [
                    "can't be blank", "is overpaid", "is forbiden", "doesn't exist"
                ]
            }
        }

    Arguments:

        errors_store: either a dict into which errors will be added or an object. If
            object we expect it to have attribute ``errors`` and that these are a dict.
            If no such attribute exists on given object, we will attach our own.
    """
    _SELF_ERRORS_KEY = "_schema"

    dest = None
    if errors_store is None:
        raise RuntimeError(
            "errors_store can't be None! errors_store must be a dict or object onto "
            "which dict can be attached!"
        )
    elif isinstance(errors_store, abc.Mapping):
        dest = errors_store
    else:
        errors_store.errors = getval(errors_store, "errors", dict())
        dest = errors_store.errors

    data = _normalize_error_data(error)

    if is_blank(data):
        return dest

    if isinstance(data, str):
        dest[_SELF_ERRORS_KEY] = getval(dest, _SELF_ERRORS_KEY, [])
        dest[_SELF_ERRORS_KEY].append(data)
        return dest

    if isinstance(data, list):
        dest[_SELF_ERRORS_KEY] = getval(dest, _SELF_ERRORS_KEY, [])
        dest[_SELF_ERRORS_KEY].extend(data)
        return dest

    for k, v in data.items():
        if is_blank(v):
            continue

        if k in dest:
            if isinstance(dest[k], list):
                if isinstance(v, str):
                    dest[k].append(v)

                elif isinstance(v, list):
                    dest[k].extend(v)

                elif isinstance(v, dict):
                    dest[k] = {_SELF_ERRORS_KEY: dest[k]}
                    add_error_to(dest[k], v)

                else:
                    dest[k].append(str(v))

            else:
                if isinstance(v, str):
                    dest[k][_SELF_ERRORS_KEY] = getval(dest[k], _SELF_ERRORS_KEY, [])
                    dest[k][_SELF_ERRORS_KEY].append(v)

                elif isinstance(v, list):
                    dest[k][_SELF_ERRORS_KEY] = getval(dest[k], _SELF_ERRORS_KEY, [])
                    dest[k][_SELF_ERRORS_KEY].extend(v)

                elif isinstance(v, dict):
                    add_error_to(dest[k], v)

                else:
                    dest[k][_SELF_ERRORS_KEY] = getval(dest[k], _SELF_ERRORS_KEY, [])
                    dest[k][_SELF_ERRORS_KEY].append(str(v))

        else:
            if isinstance(v, str):
                dest[k] = [v]

            elif isinstance(v, list):
                dest[k] = v

            elif isinstance(v, dict):
                dest[k] = dict()
                add_error_to(dest[k], v)

            else:
                dest[k] = [str(v)]

    return dest



import time

from seveno_pyutil import Stopwatch


class DescribeStopwatch(object):
    def it_provides_context_manager_for_duration_measurement(self):
        with Stopwatch() as stopwatch:
            time.sleep(1)

        assert stopwatch.duration_ms >= 1000
